﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using FunctionalMandelbrotParallel;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.UI;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Imaging;
using Windows.UI.Xaml.Navigation;

// The Blank Page item template is documented at http://go.microsoft.com/fwlink/?LinkId=234238

namespace Mandelbrot.Metro {
  /// <summary>
  /// An empty page that can be used on its own or navigated to within a Frame.
  /// </summary>
  public sealed partial class MainPage : Page {
    public MainPage() {
      this.InitializeComponent();
    }

    /// <summary>
    /// Invoked when this page is about to be displayed in a Frame.
    /// </summary>
    /// <param name="e">Event data that describes how this page was reached.  The Parameter
    /// property is typically used to configure the page.</param>
    protected override void OnNavigatedTo(NavigationEventArgs e) {
    }

    const double xstart = -2.1;
    const double xend = 1.0;
    const double ystart = -1.3;
    const double yend = 1.3;

    WriteableBitmap paintImage;
    ConcurrentQueue<PointResult> resultQueue;

    private async void DrawMandelbrot(object sender, RoutedEventArgs e) {
      int width = (int) image.ActualWidth;
      int height = (int) image.ActualHeight;
      paintImage = BitmapFactory.New(width, height);
      image.Source = paintImage;
      double xstep = (xend - xstart) / width;
      double ystep = (yend - ystart) / height;

      resultQueue = new ConcurrentQueue<PointResult>();
      var uiCancel = new CancellationTokenSource();
      Task uiUpdateTask = Task.Factory.StartNew(() => UIUpdate(uiCancel.Token), uiCancel.Token,
        TaskCreationOptions.None, TaskScheduler.FromCurrentSynchronizationContext());

      var calcInfo = new CalcInfo(xstart, xstep, ystart, ystep);
      var mainTask = Task.Factory.StartNew(() =>
        Calculator.CalcArea(width, height, calcInfo, AcceptResult));
      await mainTask;

      var waitForQueueEnd = Task.Factory.StartNew(() => {
        while (resultQueue.Count > 0)
          Task.Delay(1000);
      });
      await waitForQueueEnd;
      uiCancel.Cancel();
    }

    void AcceptResult(PointResult pointResult) {
      resultQueue.Enqueue(pointResult);
    }

    const int UPDATE_MAX = 10000;

    private void UIUpdateFromResults(int maxResults) {
      using (var paintContext = paintImage.GetBitmapContext()) {
        for (int i = 0; i < maxResults; i++) {
          PointResult pointResult;

          if (resultQueue.TryDequeue(out pointResult)) {
            paintContext.Pixels[(pointResult.Point.Y * paintContext.Width) + pointResult.Point.X] = pointResult.Color.ToArgb();
          }
          else
            break;
        }
      }
    }

    async void UIUpdate(CancellationToken token) {
      try {
        while (!(token.IsCancellationRequested)) {
          UIUpdateFromResults(UPDATE_MAX);

          await Task.Delay(600, token);
        }
        UIUpdateFromResults(resultQueue.Count);
      }
      catch (TaskCanceledException) { }
    }

  }
}